@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# Copyright: Damien Elmes <anki@ichi2.net>
+# Copyright: Ankitects Pty Ltd and contributors
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
 import pprint
@@ -11,12 +11,12 @@ import stat
 import datetime
 import copy
 import traceback
+import json
 
 from anki.lang import _, ngettext
-from anki.utils import ids2str, fieldChecksum, stripHTML, \
-    intTime, splitFields, joinFields, maxID, json
+from anki.utils import ids2str, fieldChecksum, \
+    intTime, splitFields, joinFields, maxID, devMode, stripHTMLMedia
 from anki.hooks import  runFilter, runHook
-from anki.sched import Scheduler
 from anki.models import ModelManager
 from anki.media import MediaManager
 from anki.decks import DeckManager
@@ -30,7 +30,6 @@ import anki.notes
 import anki.template
 import anki.find
 
-
 defaultConf = {
     # review options
     'activeDecks': [1],
@@ -46,10 +45,18 @@ defaultConf = {
     'sortType': "noteFld",
     'sortBackwards': False,
     'addToCur': True, # add new to currently selected deck?
+    'dayLearnFirst': False,
+    'schedVer': 2,
 }
 
+def timezoneOffset():
+    if time.localtime().tm_isdst:
+        return time.altzone//60
+    else:
+        return time.timezone//60
+
 # this is initialized by storage.Collection
-class _Collection(object):
+class _Collection:
 
     def __init__(self, db, server=False, log=False):
         self._debugLog = log
@@ -71,7 +78,8 @@ class _Collection(object):
             d = datetime.datetime(d.year, d.month, d.day)
             d += datetime.timedelta(hours=4)
             self.crt = int(time.mktime(d.timetuple()))
-        self.sched = Scheduler(self)
+        self.conf['localOffset'] = timezoneOffset()
+        self._loadScheduler()
         if not self.conf.get("newBury", False):
             self.conf['newBury'] = True
             self.setMod()
@@ -80,6 +88,49 @@ class _Collection(object):
         n = os.path.splitext(os.path.basename(self.path))[0]
         return n
 
+    # Scheduler
+    ##########################################################################
+
+    supportedSchedulerVersions = (1, 2)
+
+    def schedVer(self):
+        ver = self.conf.get("schedVer", 1)
+        if ver in self.supportedSchedulerVersions:
+            return ver
+        else:
+            raise Exception("Unsupported scheduler version")
+
+    def _loadScheduler(self):
+        ver = self.schedVer()
+        if ver == 1:
+            from anki.sched import Scheduler
+        elif ver == 2:
+            from anki.schedv2 import Scheduler
+
+        self.sched = Scheduler(self)
+
+    def changeSchedulerVer(self, ver):
+        if ver == self.schedVer():
+            return
+        if ver not in self.supportedSchedulerVersions:
+            raise Exception("Unsupported scheduler version")
+
+        self.modSchema(check=True)
+        self.clearUndo()
+
+        from anki.schedv2 import Scheduler
+        v2Sched = Scheduler(self)
+
+        if ver == 1:
+            v2Sched.moveToV1()
+        else:
+            v2Sched.moveToV2()
+
+        self.conf['schedVer'] = ver
+        self.setMod()
+
+        self._loadScheduler()
+
     # DB-related
     ##########################################################################
 
@@ -134,9 +185,10 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
         self._lastSave = time.time()
 
     def autosave(self):
-        "Save if 5 minutes has passed since last save."
+        "Save if 5 minutes has passed since last save. True if saved."
         if time.time() - self._lastSave > 300:
             self.save()
+            return True
 
     def lock(self):
         # make sure we don't accidentally bump mod time
@@ -152,7 +204,9 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
             else:
                 self.db.rollback()
             if not self.server:
+                self.db.setAutocommit(True)
                 self.db.execute("pragma journal_mode = delete")
+                self.db.setAutocommit(False)
             self.db.close()
             self.db = None
             self.media.close()
@@ -200,6 +254,7 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
         self.modSchema(check=False)
         self.ls = self.scm
         # ensure db is compacted before upload
+        self.db.setAutocommit(True)
         self.db.execute("vacuum")
         self.db.execute("analyze")
         self.close()
@@ -304,8 +359,9 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
         snids = ids2str(nids)
         have = {}
         dids = {}
-        for id, nid, ord, did, odid in self.db.execute(
-            "select id, nid, ord, did, odid from cards where nid in "+snids):
+        dues = {}
+        for id, nid, ord, did, due, odue, odid, type in self.db.execute(
+            "select id, nid, ord, did, due, odue, odid, type from cards where nid in "+snids):
             # existing cards
             if nid not in have:
                 have[nid] = {}
@@ -322,6 +378,12 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
             else:
                 # first card or multiple cards in same deck
                 dids[nid] = did
+            # save due
+            if odid != 0:
+                due = odue
+            if nid not in dues and type == 0:
+                # Add due to new card only if it's the due of a new sibling
+                dues[nid] = due
         # build cards for each note
         data = []
         ts = maxID(self.db)
@@ -333,6 +395,7 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
             model = self.models.get(mid)
             avail = self.models.availOrds(model, flds)
             did = dids.get(nid) or model['did']
+            due = dues.get(nid)
             # add any missing cards
             for t in self._tmplsFromOrds(model, avail):
                 doHave = nid in have and t['ord'] in have[nid]
@@ -343,15 +406,15 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
                         did = 1
                     # if the deck doesn't exist, use default instead
                     did = self.decks.get(did)['id']
-                    # we'd like to use the same due# as sibling cards, but we
-                    # can't retrieve that quickly, so we give it a new id
-                    # instead
+                    # use sibling due# if there is one, else use a new id
+                    if due is None:
+                        due = self.nextID("pos")
                     data.append((ts, nid, did, t['ord'],
-                                 now, usn, self.nextID("pos")))
+                                 now, usn, due))
                     ts += 1
             # note any cards that need removing
             if nid in have:
-                for ord, id in have[nid].items():
+                for ord, id in list(have[nid].items()):
                     if ord not in avail:
                         rem.append(id)
         # bulk update
@@ -363,7 +426,7 @@ insert into cards values (?,?,?,?,?,?,0,0,?,0,0,0,0,0,0,0,0,"")""",
     # type 0 - when previewing in add dialog, only non-empty
     # type 1 - when previewing edit, only existing
     # type 2 - when previewing in models dialog, all templates
-    def previewCards(self, note, type=0):
+    def previewCards(self, note, type=0, did=None):
         if type == 0:
             cms = self.findTemplates(note)
         elif type == 1:
@@ -374,19 +437,23 @@ insert into cards values (?,?,?,?,?,?,0,0,?,0,0,0,0,0,0,0,0,"")""",
             return []
         cards = []
         for template in cms:
-            cards.append(self._newCard(note, template, 1, flush=False))
+            cards.append(self._newCard(note, template, 1, flush=False, did=did))
         return cards
 
-    def _newCard(self, note, template, due, flush=True):
+    def _newCard(self, note, template, due, flush=True, did=None):
         "Create a new card."
         card = anki.cards.Card(self)
         card.nid = note.id
         card.ord = template['ord']
-        # Use template did (deck override) if valid, otherwise model did
-        if template['did'] and unicode(template['did']) in self.decks.decks:
-            card.did = template['did']
-        else:
-            card.did = note.model()['did']
+        card.did = self.db.scalar("select did from cards where nid = ? and ord = ?", card.nid, card.ord)
+        # Use template did (deck override) if valid, otherwise did in argument, otherwise model did
+        if not card.did:
+            if template['did'] and str(template['did']) in self.decks.decks:
+                card.did = template['did']
+            elif did:
+                card.did = did
+            else:
+                card.did = note.model()['did']
         # if invalid did, use default instead
         deck = self.decks.get(card.did)
         if deck['dyn']:
@@ -469,7 +536,7 @@ where c.nid = n.id and c.id in %s group by nid""" % ids2str(cids)):
             if not model:
                 # note points to invalid model
                 continue
-            r.append((stripHTML(fields[self.models.sortIdx(model)]),
+            r.append((stripHTMLMedia(fields[self.models.sortIdx(model)]),
                       fieldChecksum(fields[0]),
                       nid))
         # apply, relying on calling code to bump usn+mod
@@ -495,17 +562,18 @@ where c.nid = n.id and c.id in %s group by nid""" % ids2str(cids)):
 
     def _renderQA(self, data, qfmt=None, afmt=None):
         "Returns hash of id, question, answer."
-        # data is [cid, nid, mid, did, ord, tags, flds]
+        # data is [cid, nid, mid, did, ord, tags, flds, cardFlags]
         # unpack fields and create dict
         flist = splitFields(data[6])
         fields = {}
         model = self.models.get(data[2])
-        for (name, (idx, conf)) in self.models.fieldMap(model).items():
+        for (name, (idx, conf)) in list(self.models.fieldMap(model).items()):
             fields[name] = flist[idx]
         fields['Tags'] = data[5].strip()
         fields['Type'] = model['name']
         fields['Deck'] = self.decks.name(data[3])
         fields['Subdeck'] = fields['Deck'].split('::')[-1]
+        fields['CardFlag'] = self._flagNameFromCardFlags(data[7])
         if model['type'] == MODEL_STD:
             template = model['tmpls'][data[4]]
         else:
@@ -539,13 +607,19 @@ where c.nid = n.id and c.id in %s group by nid""" % ids2str(cids)):
         return d
 
     def _qaData(self, where=""):
-        "Return [cid, nid, mid, did, ord, tags, flds] db query"
+        "Return [cid, nid, mid, did, ord, tags, flds, cardFlags] db query"
         return self.db.execute("""
-select c.id, f.id, f.mid, c.did, c.ord, f.tags, f.flds
+select c.id, f.id, f.mid, c.did, c.ord, f.tags, f.flds, c.flags
 from cards c, notes f
 where c.nid == f.id
 %s""" % where)
 
+    def _flagNameFromCardFlags(self, flags):
+        flag = flags & 0b111
+        if not flag:
+            return ""
+        return "flag%d" % flag
+
     # Finding cards
     ##########################################################################
 
@@ -688,6 +762,7 @@ select id from notes where mid = ?) limit 1""" %
     def fixIntegrity(self):
         "Fix possible problems and rebuild caches."
         problems = []
+        curs = self.db.cursor()
         self.save()
         oldSize = os.stat(self.path)[stat.ST_SIZE]
         if self.db.scalar("pragma integrity_check") != "ok":
@@ -781,21 +856,35 @@ select id from cards where odid > 0 and did in %s""" % ids2str(dids))
         # field cache
         for m in self.models.all():
             self.updateFieldCache(self.models.nids(m))
-        # new cards can't have a due position > 32 bits
-        self.db.execute("""
-update cards set due = 1000000, mod = ?, usn = ? where due > 1000000
-and queue = 0""", intTime(), self.usn())
+        # new cards can't have a due position > 32 bits, so wrap items over
+        # 2 million back to 1 million
+        curs.execute("""
+update cards set due=1000000+due%1000000,mod=?,usn=? where due>=1000000
+and type=0""", [intTime(), self.usn()])
+        if curs.rowcount:
+            problems.append("Found %d new cards with a due number >= 1,000,000 - consider repositioning them in the Browse screen." % curs.rowcount)
         # new card position
         self.conf['nextPos'] = self.db.scalar(
             "select max(due)+1 from cards where type = 0") or 0
         # reviews should have a reasonable due #
         ids = self.db.list(
-            "select id from cards where queue = 2 and due > 10000")
+            "select id from cards where queue = 2 and due > 100000")
         if ids:
             problems.append("Reviews had incorrect due date.")
             self.db.execute(
-                "update cards set due = 0, mod = ?, usn = ? where id in %s"
-                % ids2str(ids), intTime(), self.usn())
+                "update cards set due = ?, ivl = 1, mod = ?, usn = ? where id in %s"
+                % ids2str(ids), self.sched.today, intTime(), self.usn())
+        # v2 sched had a bug that could create decimal intervals
+        curs.execute("update cards set ivl=round(ivl),due=round(due) where ivl!=round(ivl) or due!=round(due)")
+        if curs.rowcount:
+            problems.append("Fixed %d cards with v2 scheduler bug." % curs.rowcount)
+
+        curs.execute("update revlog set ivl=round(ivl),lastIvl=round(lastIvl) where ivl!=round(ivl) or lastIvl!=round(lastIvl)")
+        if curs.rowcount:
+            problems.append("Fixed %d review history entries with v2 scheduler bug." % curs.rowcount)
+        # models
+        if self.models.ensureNotEmpty():
+            problems.append("Added missing note type.")
         # and finally, optimize
         self.optimize()
         newSize = os.stat(self.path)[stat.ST_SIZE]
@@ -809,8 +898,10 @@ and queue = 0""", intTime(), self.usn())
         return ("\n".join(problems), ok)
 
     def optimize(self):
+        self.db.setAutocommit(True)
         self.db.execute("vacuum")
         self.db.execute("analyze")
+        self.db.setAutocommit(False)
         self.lock()
 
     # Logging
@@ -820,27 +911,38 @@ and queue = 0""", intTime(), self.usn())
         if not self._debugLog:
             return
         def customRepr(x):
-            if isinstance(x, basestring):
+            if isinstance(x, str):
                 return x
             return pprint.pformat(x)
         path, num, fn, y = traceback.extract_stack(
             limit=2+kwargs.get("stack", 0))[0]
-        buf = u"[%s] %s:%s(): %s" % (intTime(), os.path.basename(path), fn,
+        buf = "[%s] %s:%s(): %s" % (intTime(), os.path.basename(path), fn,
                                      ", ".join([customRepr(x) for x in args]))
-        self._logHnd.write(buf.encode("utf8") + "\n")
-        if os.environ.get("ANKIDEV"):
-            print buf
+        self._logHnd.write(buf + "\n")
+        if devMode:
+            print(buf)
 
     def _openLog(self):
         if not self._debugLog:
             return
-        lpath = re.sub("\.anki2$", ".log", self.path)
+        lpath = re.sub(r"\.anki2$", ".log", self.path)
         if os.path.exists(lpath) and os.path.getsize(lpath) > 10*1024*1024:
             lpath2 = lpath + ".old"
             if os.path.exists(lpath2):
                 os.unlink(lpath2)
             os.rename(lpath, lpath2)
-        self._logHnd = open(lpath, "ab")
+        self._logHnd = open(lpath, "a", encoding="utf8")
 
     def _closeLog(self):
+        if not self._debugLog:
+            return
+        self._logHnd.close()
         self._logHnd = None
+
+    # Card Flags
+    ##########################################################################
+
+    def setUserFlag(self, flag, cids):
+        assert 0 <= flag <= 7
+        self.db.execute("update cards set flags = (flags & ~?) | ?, usn=?, mod=? where id in %s" %
+                        ids2str(cids), 0b111, flag, self.usn(), intTime())
