@@ -1,13 +1,14 @@
 # -*- coding: utf-8 -*-
-# Copyright: Damien Elmes <anki@ichi2.net>
+# Copyright: Ankitects Pty Ltd and contributors
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
-import copy, re
+import copy, re, json
 from anki.utils import intTime, joinFields, splitFields, ids2str,\
-    checksum, json
+    checksum
 from anki.lang import _
 from anki.consts import *
 from anki.hooks import runHook
+import time
 
 # Models
 ##########################################################################
@@ -20,6 +21,7 @@ defaultModel = {
     'latexPre': """\
 \\documentclass[12pt]{article}
 \\special{papersize=3in,5in}
+\\usepackage[utf8]{inputenc}
 \\usepackage{amssymb,amsmath}
 \\pagestyle{empty}
 \\setlength{\\parindent}{0in}
@@ -67,7 +69,7 @@ defaultTemplate = {
     #'bsize': 12,
 }
 
-class ModelManager(object):
+class ModelManager:
 
     # Saving/loading registry
     #############################################################
@@ -94,10 +96,17 @@ class ModelManager(object):
     def flush(self):
         "Flush the registry if any models were changed."
         if self.changed:
+            self.ensureNotEmpty()
             self.col.db.execute("update col set models = ?",
                                  json.dumps(self.models))
             self.changed = False
 
+    def ensureNotEmpty(self):
+        if not self.models:
+            from anki.stdmodels import addBasicModel
+            addBasicModel(self.col)
+            return True
+
     # Retrieving and creating models
     #############################################################
 
@@ -106,7 +115,7 @@ class ModelManager(object):
         m = self.get(self.col.decks.current().get('mid'))
         if not forDeck or not m:
             m = self.get(self.col.conf['curModel'])
-        return m or self.models.values()[0]
+        return m or list(self.models.values())[0]
 
     def setCurrent(self, m):
         self.col.conf['curModel'] = m['id']
@@ -120,14 +129,14 @@ class ModelManager(object):
 
     def all(self):
         "Get all models."
-        return self.models.values()
+        return list(self.models.values())
 
     def allNames(self):
         return [m['name'] for m in self.all()]
 
     def byName(self, name):
         "Get model with NAME."
-        for m in self.models.values():
+        for m in list(self.models.values()):
             if m['name'] == name:
                 return m
 
@@ -145,7 +154,7 @@ class ModelManager(object):
 
     def rem(self, m):
         "Delete model, and all its cards/notes."
-        self.col.modSchema()
+        self.col.modSchema(check=True)
         current = self.current()['id'] == m['id']
         # delete notes/cards
         self.col.remCards(self.col.db.list("""
@@ -156,7 +165,7 @@ select id from cards where nid in (select id from notes where mid = ?)""",
         self.save()
         # GUI should ensure last model is not deleted
         if current:
-            self.setCurrent(self.models.values()[0])
+            self.setCurrent(list(self.models.values())[0])
 
     def add(self, m):
         self._setID(m)
@@ -164,8 +173,15 @@ select id from cards where nid in (select id from notes where mid = ?)""",
         self.setCurrent(m)
         self.save(m)
 
+    def ensureNameUnique(self, m):
+        for mcur in self.all():
+            if (mcur['name'] == m['name'] and mcur['id'] != m['id']):
+                m['name'] += "-" + checksum(str(time.time()))[:5]
+                break
+
     def update(self, m):
         "Add or update an existing model. Used for syncing and merging."
+        self.ensureNameUnique(m)
         self.models[str(m['id'])] = m
         # mark registry changed, but don't bump mod time
         self.save()
@@ -181,7 +197,7 @@ select id from cards where nid in (select id from notes where mid = ?)""",
         return str(id) in self.models
 
     def ids(self):
-        return self.models.keys()
+        return list(self.models.keys())
 
     # Tools
     ##################################################
@@ -196,6 +212,11 @@ select id from cards where nid in (select id from notes where mid = ?)""",
         return self.col.db.scalar(
             "select count() from notes where mid = ?", m['id'])
 
+    def tmplUseCount(self, m, ord):
+        return self.col.db.scalar("""
+select count() from cards, notes where cards.nid = notes.id
+and notes.mid = ? and cards.ord = ?""", m['id'], ord)
+
     # Copying
     ##################################################
 
@@ -210,6 +231,7 @@ select id from cards where nid in (select id from notes where mid = ?)""",
     ##################################################
 
     def newField(self, name):
+        assert(isinstance(name, str))
         f = defaultField.copy()
         f['name'] = name
         return f
@@ -225,8 +247,8 @@ select id from cards where nid in (select id from notes where mid = ?)""",
         return m['sortf']
 
     def setSortIdx(self, m, idx):
-        assert idx >= 0 and idx < len(m['flds'])
-        self.col.modSchema()
+        assert 0 <= idx < len(m['flds'])
+        self.col.modSchema(check=True)
         m['sortf'] = idx
         self.col.updateFieldCache(self.nids(m))
         self.save(m)
@@ -234,7 +256,7 @@ select id from cards where nid in (select id from notes where mid = ?)""",
     def addField(self, m, field):
         # only mod schema if model isn't new
         if m['id']:
-            self.col.modSchema()
+            self.col.modSchema(check=True)
         m['flds'].append(field)
         self._updateFieldOrds(m)
         self.save(m)
@@ -244,24 +266,30 @@ select id from cards where nid in (select id from notes where mid = ?)""",
         self._transformFields(m, add)
 
     def remField(self, m, field):
-        self.col.modSchema()
+        self.col.modSchema(check=True)
+        # save old sort field
+        sortFldName = m['flds'][m['sortf']]['name']
         idx = m['flds'].index(field)
         m['flds'].remove(field)
-        if m['sortf'] >= len(m['flds']):
-            m['sortf'] -= 1
+        # restore old sort field if possible, or revert to first field
+        m['sortf'] = 0
+        for c, f in enumerate(m['flds']):
+            if f['name'] == sortFldName:
+                m['sortf'] = c
+                break
         self._updateFieldOrds(m)
         def delete(fields):
             del fields[idx]
             return fields
         self._transformFields(m, delete)
-        if idx == self.sortIdx(m):
-            # need to rebuild
+        if m['flds'][m['sortf']]['name'] != sortFldName:
+            # need to rebuild sort field
             self.col.updateFieldCache(self.nids(m))
         # saves
         self.renameField(m, field, None)
 
     def moveField(self, m, field, idx):
-        self.col.modSchema()
+        self.col.modSchema(check=True)
         oldidx = m['flds'].index(field)
         if oldidx == idx:
             return
@@ -282,11 +310,11 @@ select id from cards where nid in (select id from notes where mid = ?)""",
         self._transformFields(m, move)
 
     def renameField(self, m, field, newName):
-        self.col.modSchema()
-        pat = r'{{([:#^/]|[^:#/^}][^:}]*?:|)%s}}'
+        self.col.modSchema(check=True)
+        pat = r'{{([^{}]*)([:#^/]|[^:#/^}][^:}]*?:|)%s}}'
         def wrap(txt):
             def repl(match):
-                return '{{' + match.group(1) + txt +  '}}'
+                return '{{' + match.group(1) + match.group(2) + txt +  '}}'
             return repl
         for t in m['tmpls']:
             for fmt in ('qfmt', 'afmt'):
@@ -326,7 +354,7 @@ select id from cards where nid in (select id from notes where mid = ?)""",
     def addTemplate(self, m, template):
         "Note: should col.genCards() afterwards."
         if m['id']:
-            self.col.modSchema()
+            self.col.modSchema(check=True)
         m['tmpls'].append(template)
         self._updateTemplOrds(m)
         self.save(m)
@@ -349,7 +377,7 @@ having count() < 2
 limit 1""" % ids2str(cids)):
             return False
         # ok to proceed; remove cards
-        self.col.modSchema()
+        self.col.modSchema(check=True)
         self.col.remCards(cids)
         # shift ordinals
         self.col.db.execute("""
@@ -393,7 +421,7 @@ select id from notes where mid = ?)""" % " ".join(map),
     # - newModel should be self if model is not changing
 
     def change(self, m, nids, newModel, fmap, cmap):
-        self.col.modSchema()
+        self.col.modSchema(check=True)
         assert newModel['id'] == m['id'] or (fmap and cmap)
         if fmap:
             self._changeNotes(nids, newModel, fmap)
@@ -408,7 +436,7 @@ select id from notes where mid = ?)""" % " ".join(map),
             "select id, flds from notes where id in "+ids2str(nids)):
             newflds = {}
             flds = splitFields(flds)
-            for old, new in map.items():
+            for old, new in list(map.items()):
                 newflds[new] = flds[old]
             flds = []
             for c in range(nfields):
@@ -457,8 +485,6 @@ select id from notes where mid = ?)""" % " ".join(map),
             s += f['name']
         for t in m['tmpls']:
             s += t['name']
-            s += t['qfmt']
-            s += t['afmt']
         return checksum(s)
 
     # Required field/text cache
@@ -481,9 +507,9 @@ select id from notes where mid = ?)""" % " ".join(map),
         for f in flds:
             a.append("ankiflag")
             b.append("")
-        data = [1, 1, m['id'], 1, t['ord'], "", joinFields(a)]
+        data = [1, 1, m['id'], 1, t['ord'], "", joinFields(a), 0]
         full = self.col._renderQA(data)['q']
-        data = [1, 1, m['id'], 1, t['ord'], "", joinFields(b)]
+        data = [1, 1, m['id'], 1, t['ord'], "", joinFields(b), 0]
         empty = self.col._renderQA(data)['q']
         # if full and empty are the same, the template is invalid and there is
         # no way to satisfy it
@@ -550,14 +576,14 @@ select id from notes where mid = ?)""" % " ".join(map),
         sflds = splitFields(flds)
         map = self.fieldMap(m)
         ords = set()
-        matches = re.findall("{{cloze:(.+?)}}", m['tmpls'][0]['qfmt'])
+        matches = re.findall("{{[^}]*?cloze:(?:[^}]?:)*(.+?)}}", m['tmpls'][0]['qfmt'])
         matches += re.findall("<%cloze:(.+?)%>", m['tmpls'][0]['qfmt'])
         for fname in matches:
             if fname not in map:
                 continue
             ord = map[fname][0]
             ords.update([int(m)-1 for m in re.findall(
-                "{{c(\d+)::.+?}}", sflds[ord])])
+                r"(?s){{c(\d+)::.+?}}", sflds[ord])])
         if -1 in ords:
             ords.remove(-1)
         if not ords and allowEmpty:
