@@ -1,13 +1,11 @@
 # -*- coding: utf-8 -*-
-# Copyright: Damien Elmes <anki@ichi2.net>
+# Copyright: Ankitects Pty Ltd and contributors
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
-from __future__ import division
 import time
 import datetime
 import json
 
-import anki.js
 from anki.utils import fmtTimeSpan, ids2str
 from anki.lang import _, ngettext
 
@@ -15,7 +13,7 @@ from anki.lang import _, ngettext
 # Card stats
 ##########################################################################
 
-class CardStats(object):
+class CardStats:
 
     def __init__(self, col, card):
         self.col = col
@@ -23,6 +21,7 @@ class CardStats(object):
 
     def report(self):
         c = self.card
+        # pylint: disable=unnecessary-lambda
         fmt = lambda x, **kwargs: fmtTimeSpan(x, short=True, **kwargs)
         self.txt = "<table width=100%>"
         self.addLine(_("Added"), self.date(c.id/1000))
@@ -99,7 +98,7 @@ colTime = "#770"
 colUnseen = "#000"
 colSusp = "#ff0"
 
-class CollectionStats(object):
+class CollectionStats:
 
     def __init__(self, col):
         self.col = col
@@ -109,27 +108,34 @@ class CollectionStats(object):
         self.height = 200
         self.wholeCollection = False
 
+    # assumes jquery & plot are available in document
     def report(self, type=0):
         # 0=days, 1=weeks, 2=months
         self.type = type
-        from statsbg import bg
+        from .statsbg import bg
         txt = self.css % bg
-        txt += self.todayStats()
-        txt += self.dueGraph()
-        txt += self.repsGraph()
-        txt += self.ivlGraph()
-        txt += self.hourGraph()
-        txt += self.easeGraph()
-        txt += self.cardGraph()
-        txt += self.footer()
-        return "<script>%s\n</script><center>%s</center>" % (
-            anki.js.jquery+anki.js.plot, txt)
+        txt += self._section(self.todayStats())
+        txt += self._section(self.dueGraph())
+        txt += self.repsGraphs()
+        txt += self._section(self.introductionGraph())
+        txt += self._section(self.ivlGraph())
+        txt += self._section(self.hourGraph())
+        txt += self._section(self.easeGraph())
+        txt += self._section(self.cardGraph())
+        txt += self._section(self.footer())
+        return "<center>%s</center>" % txt
+
+    def _section(self, txt):
+        return "<div class=section>%s</div>" % txt
 
     css = """
 <style>
 h1 { margin-bottom: 0; margin-top: 1em; }
 .pielabel { text-align:center; padding:0px; color:white; }
 body {background-image: url(data:image/png;base64,%s); }
+@media print {
+    .section { page-break-inside: avoid; padding-top: 5mm; }
+}
 </style>
 """
 
@@ -159,41 +165,57 @@ from revlog where id > ? """+lim, (self.col.sched.dayCutoff-86400)*1000)
         filt = filt or 0
         # studied
         def bold(s):
-            return "<b>"+unicode(s)+"</b>"
+            return "<b>"+str(s)+"</b>"
         msgp1 = ngettext("<!--studied-->%d card", "<!--studied-->%d cards", cards) % cards
-        b += _("Studied %(a)s in %(b)s today.") % dict(
-            a=bold(msgp1), b=bold(fmtTimeSpan(thetime, unit=1)))
-        # again/pass count
-        b += "<br>" + _("Again count: %s") % bold(failed)
         if cards:
-            b += " " + _("(%s correct)") % bold(
-                "%0.1f%%" %((1-failed/float(cards))*100))
-        # type breakdown
-        b += "<br>"
-        b += (_("Learn: %(a)s, Review: %(b)s, Relearn: %(c)s, Filtered: %(d)s")
-              % dict(a=bold(lrn), b=bold(rev), c=bold(relrn), d=bold(filt)))
-        # mature today
-        mcnt, msum = self.col.db.first("""
-select count(), sum(case when ease = 1 then 0 else 1 end) from revlog
-where lastIvl >= 21 and id > ?"""+lim, (self.col.sched.dayCutoff-86400)*1000)
-        b += "<br>"
-        if mcnt:
-            b += _("Correct answers on mature cards: %(a)d/%(b)d (%(c).1f%%)") % dict(
-                a=msum, b=mcnt, c=(msum / float(mcnt) * 100))
+            b += _("Studied %(a)s %(b)s today (%(secs).1fs/card)") % dict(
+                a=bold(msgp1), b=bold(fmtTimeSpan(thetime, unit=1, inTime=True)),
+                secs=thetime/cards
+            )
+            # again/pass count
+            b += "<br>" + _("Again count: %s") % bold(failed)
+            if cards:
+                b += " " + _("(%s correct)") % bold(
+                    "%0.1f%%" %((1-failed/float(cards))*100))
+            # type breakdown
+            b += "<br>"
+            b += (_("Learn: %(a)s, Review: %(b)s, Relearn: %(c)s, Filtered: %(d)s")
+                  % dict(a=bold(lrn), b=bold(rev), c=bold(relrn), d=bold(filt)))
+            # mature today
+            mcnt, msum = self.col.db.first("""
+    select count(), sum(case when ease = 1 then 0 else 1 end) from revlog
+    where lastIvl >= 21 and id > ?"""+lim, (self.col.sched.dayCutoff-86400)*1000)
+            b += "<br>"
+            if mcnt:
+                b += _("Correct answers on mature cards: %(a)d/%(b)d (%(c).1f%%)") % dict(
+                    a=msum, b=mcnt, c=(msum / float(mcnt) * 100))
+            else:
+                b += _("No mature cards were studied today.")
         else:
-            b += _("No mature cards were studied today.")
+            b += _("No cards have been studied today.")
         return b
 
     # Due and cumulative due
     ######################################################################
 
-    def dueGraph(self):
+    def get_start_end_chunk(self, by='review'):
+        start = 0
         if self.type == 0:
-            start = 0; end = 31; chunk = 1;
+            end, chunk = 31, 1
         elif self.type == 1:
-            start = 0; end = 52; chunk = 7
+            end, chunk = 52, 7
         elif self.type == 2:
-            start = 0; end = None; chunk = 30
+            end = None
+            if self._deckAge(by) <= 100:
+                chunk = 1
+            elif self._deckAge(by) <= 700:
+                chunk = 7
+            else:
+                chunk = 31
+        return start, end, chunk
+
+    def dueGraph(self):
+        start, end, chunk = self.get_start_end_chunk()
         d = self._due(start, end, chunk)
         yng = []
         mtr = []
@@ -218,10 +240,13 @@ where lastIvl >= 21 and id > ?"""+lim, (self.col.sched.dayCutoff-86400)*1000)
         xaxis = dict(tickDecimals=0, min=-0.5)
         if end is not None:
             xaxis['max'] = end-0.5
-        txt += self._graph(id="due", data=data,
-                           ylabel2=_("Cumulative Cards"), conf=dict(
-                xaxis=xaxis, yaxes=[dict(min=0), dict(
-                    min=0, tickDecimals=0, position="right")]))
+        txt += self._graph(
+            id="due", data=data, xunit=chunk, ylabel2=_("Cumulative Cards"),
+            conf=dict(
+                xaxis=xaxis, yaxes=[
+                    dict(min=0), dict(min=0, tickDecimals=0, position="right")]
+            ),
+        )
         txt += self._dueInfo(tot, len(totd)*chunk)
         return txt
 
@@ -254,49 +279,71 @@ group by day order by day""" % (self._limit(), lim),
                             today=self.col.sched.today,
                             chunk=chunk)
 
-    # Reps and time spent
+    # Added, reps and time spent
     ######################################################################
 
-    def repsGraph(self):
-        if self.type == 0:
-            days = 30; chunk = 1
-        elif self.type == 1:
-            days = 52; chunk = 7
-        else:
-            days = None; chunk = 30
-        return self._repsGraph(self._done(days, chunk),
-                               days,
-                               _("Review Count"),
-                               _("Review Time"))
-
-    def _repsGraph(self, data, days, reptitle, timetitle):
+    def introductionGraph(self):
+        start, days, chunk = self.get_start_end_chunk()
+        data = self._added(days, chunk)
         if not data:
             return ""
-        d = data
         conf = dict(
             xaxis=dict(tickDecimals=0, max=0.5),
-            yaxes=[dict(min=0), dict(position="right",min=0)])
+            yaxes=[dict(min=0), dict(position="right", min=0)])
         if days is not None:
+            # pylint: disable=invalid-unary-operand-type
             conf['xaxis']['min'] = -days+0.5
         def plot(id, data, ylabel, ylabel2):
             return self._graph(
-                id, data=data, conf=conf, ylabel=ylabel, ylabel2=ylabel2)
+                id, data=data, conf=conf, xunit=chunk, ylabel=ylabel, ylabel2=ylabel2)
+        # graph
+        repdata, repsum = self._splitRepData(data, ((1, colLearn, ""),))
+        txt = self._title(
+            _("Added"), _("The number of new cards you have added."))
+        txt += plot("intro", repdata, ylabel=_("Cards"), ylabel2=_("Cumulative Cards"))
+        # total and per day average
+        tot = sum([i[1] for i in data])
+        period = self._periodDays()
+        if not period:
+            # base off date of earliest added card
+            period = self._deckAge('add')
+        i = []
+        self._line(i, _("Total"), ngettext("%d card", "%d cards", tot) % tot)
+        self._line(i, _("Average"), self._avgDay(tot, period, _("cards")))
+        txt += self._lineTbl(i)
+
+        return txt
+
+    def repsGraphs(self):
+        start, days, chunk = self.get_start_end_chunk()
+        data = self._done(days, chunk)
+        if not data:
+            return ""
+        conf = dict(
+            xaxis=dict(tickDecimals=0, max=0.5),
+            yaxes=[dict(min=0), dict(position="right", min=0)])
+        if days is not None:
+            # pylint: disable=invalid-unary-operand-type
+            conf['xaxis']['min'] = -days+0.5
+        def plot(id, data, ylabel, ylabel2):
+            return self._graph(
+                id, data=data, conf=conf, xunit=chunk, ylabel=ylabel, ylabel2=ylabel2)
         # reps
-        (repdata, repsum) = self._splitRepData(d, (
+        (repdata, repsum) = self._splitRepData(data, (
             (3, colMature, _("Mature")),
             (2, colYoung, _("Young")),
             (4, colRelearn, _("Relearn")),
             (1, colLearn, _("Learn")),
             (5, colCram, _("Cram"))))
-        txt = self._title(
-            reptitle, _("The number of questions you have answered."))
-        txt += plot("reps", repdata, ylabel=_("Answers"), ylabel2=_(
+        txt1 = self._title(
+            _("Review Count"), _("The number of questions you have answered."))
+        txt1 += plot("reps", repdata, ylabel=_("Answers"), ylabel2=_(
             "Cumulative Answers"))
         (daysStud, fstDay) = self._daysStudied()
         rep, tot = self._ansInfo(repsum, daysStud, fstDay, _("reviews"))
-        txt += rep
+        txt1 += rep
         # time
-        (timdata, timsum) = self._splitRepData(d, (
+        (timdata, timsum) = self._splitRepData(data, (
             (8, colMature, _("Mature")),
             (7, colYoung, _("Young")),
             (9, colRelearn, _("Relearn")),
@@ -308,12 +355,12 @@ group by day order by day""" % (self._limit(), lim),
         else:
             t = _("Hours")
             convHours = True
-        txt += self._title(timetitle, _("The time taken to answer the questions."))
-        txt += plot("time", timdata, ylabel=t, ylabel2=_("Cumulative %s") % t)
+        txt2 = self._title(_("Review Time"), _("The time taken to answer the questions."))
+        txt2 += plot("time", timdata, ylabel=t, ylabel2=_("Cumulative %s") % t)
         rep, tot2 = self._ansInfo(
             timsum, daysStud, fstDay, _("minutes"), convHours, total=tot)
-        txt += rep
-        return txt
+        txt2 += rep
+        return self._section(txt1) + self._section(txt2)
 
     def _ansInfo(self, totd, studied, first, unit, convHours=False, total=None):
         if not totd:
@@ -322,15 +369,7 @@ group by day order by day""" % (self._limit(), lim),
         period = self._periodDays()
         if not period:
             # base off earliest repetition date
-            lim = self._revlogLimit()
-            if lim:
-                lim = " where " + lim
-            t = self.col.db.scalar("select id from revlog %s order by id limit 1" % lim)
-            if not t:
-                period = 1
-            else:
-                period = max(
-                    1, int(1+((self.col.sched.dayCutoff - (t/1000)) / 86400)))
+            period = self._deckAge('review')
         i = []
         self._line(i, _("Days studied"),
                    _("<b>%(pct)d%%</b> (%(x)s of %(y)s)") % dict(
@@ -340,6 +379,7 @@ group by day order by day""" % (self._limit(), lim),
             tunit = _("hours")
         else:
             tunit = unit
+        #T: unit: can be hours, minutes, reviews... tot: the number of unit.
         self._line(i, _("Total"), _("%(tot)s %(unit)s") % dict(
             unit=tunit, tot=int(tot)))
         if convHours:
@@ -353,10 +393,15 @@ group by day order by day""" % (self._limit(), lim),
                 tot, period, unit))
         if total and tot:
             perMin = total / float(tot)
-            perMin = ngettext("%d card/minute", "%d cards/minute", perMin) % perMin
+            perMin = round(perMin, 1)
+            # don't round down to zero
+            if perMin < 0.1:
+                text = _("less than 0.1 cards/minute")
+            else:
+                text = _("%.01f cards/minute") % perMin
             self._line(
                 i, _("Average answer time"),
-                _("%(a)0.1fs (%(b)s)") % dict(a=(tot*60)/total, b=perMin))
+                _("%(a)0.1fs (%(b)s)") % dict(a=(tot*60)/total, b=text))
         return self._lineTbl(i), int(tot)
 
     def _splitRepData(self, data, spec):
@@ -368,7 +413,6 @@ group by day order by day""" % (self._limit(), lim),
         for (n, col, lab) in spec:
             totcnt[n] = 0
             totd[n] = []
-        sum = []
         for row in data:
             for (n, col, lab) in spec:
                 if n not in sep:
@@ -389,6 +433,27 @@ group by day order by day""" % (self._limit(), lim),
                 bars={'show': False}, lines=dict(show=True), stack=-n))
         return (ret, alltot)
 
+    def _added(self, num=7, chunk=1):
+        lims = []
+        if num is not None:
+            lims.append("id > %d" % (
+                (self.col.sched.dayCutoff-(num*chunk*86400))*1000))
+        lims.append("did in %s" % self._limit())
+        if lims:
+            lim = "where " + " and ".join(lims)
+        else:
+            lim = ""
+        if self.type == 0:
+            tf = 60.0 # minutes
+        else:
+            tf = 3600.0 # hours
+        return self.col.db.all("""
+select
+(cast((id/1000.0 - :cut) / 86400.0 as int))/:chunk as day,
+count(id)
+from cards %s
+group by day order by day""" % lim, cut=self.col.sched.dayCutoff,tf=tf, chunk=chunk)
+
     def _done(self, num=7, chunk=1):
         lims = []
         if num is not None:
@@ -450,7 +515,7 @@ group by day order by day)""" % lim,
     ######################################################################
 
     def ivlGraph(self):
-        (ivls, all, avg, max_) = self._ivls()
+        (ivls, all, avg, max_), chunk = self._ivls()
         tot = 0
         totd = []
         if not ivls or not all:
@@ -466,7 +531,7 @@ group by day order by day)""" % lim,
             ivlmax = max(5, ivls[-1][0])
         txt = self._title(_("Intervals"),
                           _("Delays until reviews are shown again."))
-        txt += self._graph(id="ivl", ylabel2=_("Percentage"), data=[
+        txt += self._graph(id="ivl", ylabel2=_("Percentage"), xunit=chunk, data=[
             dict(data=ivls, color=colIvl),
             dict(data=totd, color=colCum, yaxis=2,
              bars={'show': False}, lines=dict(show=True), stack=False)
@@ -479,12 +544,8 @@ group by day order by day)""" % lim,
         return txt + self._lineTbl(i)
 
     def _ivls(self):
-        if self.type == 0:
-            chunk = 1; lim = " and grp <= 30"
-        elif self.type == 1:
-            chunk = 7; lim = " and grp <= 52"
-        else:
-            chunk = 30; lim = ""
+        start, end, chunk = self.get_start_end_chunk()
+        lim = "and grp <= %d" % end if end else ""
         data = [self.col.db.all("""
 select ivl / :chunk as grp, count() from cards
 where did in %s and queue = 2 %s
@@ -492,7 +553,7 @@ group by grp
 order by grp""" % (self._limit(), lim), chunk=chunk)]
         return data + list(self.col.db.first("""
 select count(), avg(ivl), max(ivl) from cards where did in %s and queue = 2""" %
-                                         self._limit()))
+                                         self._limit())), chunk
 
     # Eases
     ######################################################################
@@ -511,16 +572,18 @@ select count(), avg(ivl), max(ivl) from cards where did in %s and queue = 2""" %
                 ease += 10
             n = types[type]
             d[n].append((ease, cnt))
-        ticks = [[1,1],[2,2],[3,3],
+        ticks = [[1,1],[2,2],[3,3], # [4,4]
                  [6,1],[7,2],[8,3],[9,4],
                  [11, 1],[12,2],[13,3],[14,4]]
+        if self.col.schedVer() != 1:
+            ticks.insert(3, [4,4])
         txt = self._title(_("Answer Buttons"),
                           _("The number of times you have pressed each button."))
         txt += self._graph(id="ease", data=[
             dict(data=d['lrn'], color=colLearn, label=_("Learning")),
             dict(data=d['yng'], color=colYoung, label=_("Young")),
             dict(data=d['mtr'], color=colMature, label=_("Mature")),
-            ], type="barsLine", conf=dict(
+            ], type="bars", conf=dict(
                 xaxis=dict(ticks=ticks, min=0, max=15)),
             ylabel=_("Answers"))
         txt += self._easeInfo(eases)
@@ -554,12 +617,7 @@ select count(), avg(ivl), max(ivl) from cards where did in %s and queue = 2""" %
         lim = self._revlogLimit()
         if lim:
             lims.append(lim)
-        if self.type == 0:
-            days = 30
-        elif self.type == 1:
-            days = 365
-        else:
-            days = None
+        days = self._periodDays()
         if days is not None:
             lims.append("id > %d" % (
                 (self.col.sched.dayCutoff-(days*86400))*1000))
@@ -567,14 +625,18 @@ select count(), avg(ivl), max(ivl) from cards where did in %s and queue = 2""" %
             lim = "where " + " and ".join(lims)
         else:
             lim = ""
+        if self.col.schedVer() == 1:
+            ease4repl = "3"
+        else:
+            ease4repl = "ease"
         return self.col.db.all("""
 select (case
 when type in (0,2) then 0
 when lastIvl < 21 then 1
 else 2 end) as thetype,
-(case when type in (0,2) and ease = 4 then 3 else ease end), count() from revlog %s
+(case when type in (0,2) and ease = 4 then %s else ease end), count() from revlog %s
 group by thetype, ease
-order by thetype, ease""" % lim)
+order by thetype, ease""" % (ease4repl, lim))
 
     # Hourly retention
     ######################################################################
@@ -630,7 +692,11 @@ order by thetype, ease""" % lim)
         lim = self._revlogLimit()
         if lim:
             lim = " and " + lim
-        sd = datetime.datetime.fromtimestamp(self.col.crt)
+        if self.col.schedVer() == 1:
+            sd = datetime.datetime.fromtimestamp(self.col.crt)
+            rolloverHour = sd.hour
+        else:
+            rolloverHour = self.col.conf.get("rollover", 4)
         pd = self._periodDays()
         if pd:
             lim += " and id > %d" % ((self.col.sched.dayCutoff-(86400*pd))*1000)
@@ -642,7 +708,7 @@ cast(count() as float) * 100,
 count()
 from revlog where type in (0,1,2) %s
 group by hour having count() > 30 order by hour""" % lim,
-                            cut=self.col.sched.dayCutoff-(sd.hour*3600))
+                            cut=self.col.sched.dayCutoff-(rolloverHour*3600))
 
     # Cards
     ######################################################################
@@ -655,7 +721,7 @@ group by hour having count() > 30 order by hour""" % lim,
             (_("Mature"), colMature),
             (_("Young+Learn"), colYoung),
             (_("Unseen"), colUnseen),
-            (_("Suspended"), colSusp))):
+            (_("Suspended+Buried"), colSusp))):
             d.append(dict(data=div[c], label="%s: %s" % (t, div[c]), color=col))
         # text data
         i = []
@@ -673,7 +739,7 @@ where did in %s """ % self._limit())
         info += _('''\
 A card's <i>ease</i> is the size of the next interval \
 when you answer "good" on a review.''')
-        txt = self._title(_("Cards Types"),
+        txt = self._title(_("Card Types"),
                           _("The division of cards in your deck(s)."))
         txt += "<table width=%d><tr><td>%s</td><td>%s</td></table>" % (
             self.width,
@@ -682,6 +748,7 @@ when you answer "good" on a review.''')
         return txt
 
     def _line(self, i, a, b, bold=True):
+        #T: Symbols separating first and second column in a statistics table. Eg in "Total:    3 reviews".
         colon = _(":")
         if bold:
             i.append(("<tr><td width=200 align=right>%s%s</td><td><b>%s</b></td></tr>") % (a,colon,b))
@@ -705,7 +772,7 @@ select
 sum(case when queue=2 and ivl >= 21 then 1 else 0 end), -- mtr
 sum(case when queue in (1,3) or (queue=2 and ivl < 21) then 1 else 0 end), -- yng/lrn
 sum(case when queue=0 then 1 else 0 end), -- new
-sum(case when queue=-1 then 1 else 0 end) -- susp
+sum(case when queue<0 then 1 else 0 end) -- susp
 from cards where did in %s""" % self._limit())
 
     # Footer
@@ -731,8 +798,10 @@ from cards where did in %s""" % self._limit())
     # Tools
     ######################################################################
 
-    def _graph(self, id, data, conf={},
-               type="bars", ylabel=_("Cards"), timeTicks=True, ylabel2=""):
+    def _graph(self, id, data, conf=None,
+               type="bars", xunit=1, ylabel=_("Cards"), ylabel2=""):
+        if conf is None:
+            conf = {}
         # display settings
         if type == "pie":
             conf['legend'] = {'container': "#%sLegend" % id, 'noColumns':2}
@@ -744,8 +813,16 @@ from cards where did in %s""" % self._limit())
         conf['yaxis']['labelWidth'] = 40
         if 'xaxis' not in conf:
             conf['xaxis'] = {}
-        if timeTicks:
-            conf['timeTicks'] = (_("d"), _("w"), _("mo"))[self.type]
+        if xunit is None:
+            conf['timeTicks'] = False
+        else:
+            #T: abbreviation of day
+            d = _("d")
+            #T: abbreviation of week
+            w = _("w")
+            #T: abbreviation of month
+            mo = _("mo")
+            conf['timeTicks'] = {1: d, 7: w, 31: mo}[xunit]
         # types
         width = self.width
         height = self.height
@@ -753,6 +830,7 @@ from cards where did in %s""" % self._limit())
             conf['series']['bars'] = dict(
                 show=True, barWidth=0.8, align="center", fill=0.7, lineWidth=0)
         elif type == "barsLine":
+            print("deprecated - use 'bars' instead")
             conf['series']['bars'] = dict(
                 show=True, barWidth=0.8, align="center", fill=0.7, lineWidth=3)
         elif type == "fill":
@@ -773,8 +851,6 @@ from cards where did in %s""" % self._limit())
                         opacity=0.5,
                         color="#000"
                     )))
-
-            #conf['legend'] = dict(show=False)
         return (
 """
 <table cellpadding=0 cellspacing=10>
@@ -805,8 +881,12 @@ $(function () {
         }
     }
     conf.yaxis.minTickSize = 1;
+    // prevent ticks from having decimals (use whole numbers instead)
+    conf.yaxis.tickDecimals = 0;
     conf.yaxis.tickFormatter = function (val, axis) {
-            return val.toFixed(0);
+            // Just in case we get ticks with decimals, render to one decimal position.  If it's
+            // a whole number then render without any decimal (i.e. without the trailing .0).
+            return val === Math.round(val) ? val.toFixed(0) : val.toFixed(1);
     }
     if (conf.series.pie) {
         conf.series.pie.label.formatter = function(label, series){
@@ -834,13 +914,27 @@ $(function () {
     def _title(self, title, subtitle=""):
         return '<h1>%s</h1>%s' % (title, subtitle)
 
-    def _periodDays(self):
-        if self.type == 0:
-            return 30
-        elif self.type == 1:
-            return 365
+    def _deckAge(self, by):
+        lim = self._revlogLimit()
+        if lim:
+            lim = " where " + lim
+        if by == 'review':
+            t = self.col.db.scalar("select id from revlog %s order by id limit 1" % lim)
+        elif by == 'add':
+            lim = "where did in %s" % ids2str(self.col.decks.active())
+            t = self.col.db.scalar("select id from cards %s order by id limit 1" % lim)
+        if not t:
+            period = 1
         else:
+            period = max(
+                1, int(1+((self.col.sched.dayCutoff - (t/1000)) / 86400)))
+        return period
+
+    def _periodDays(self):
+        start, end, chunk = self.get_start_end_chunk()
+        if end is None:
             return None
+        return end * chunk
 
     def _avgDay(self, tot, num, unit):
         vals = []
