@@ -1,9 +1,11 @@
-# Copyright: Damien Elmes <anki@ichi2.net>
+# Copyright: Ankitects Pty Ltd and contributors
 # -*- coding: utf-8 -*-
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
 import time
 from aqt.qt import *
+import aqt.forms
+from anki.lang import _
 
 # fixme: if mw->subwindow opens a progress dialog with mw as the parent, mw
 # gets raised on finish on compiz. perhaps we should be using the progress
@@ -12,7 +14,7 @@ from aqt.qt import *
 # Progress info
 ##########################################################################
 
-class ProgressManager(object):
+class ProgressManager:
 
     def __init__(self, mw):
         self.mw = mw
@@ -29,11 +31,7 @@ class ProgressManager(object):
         "Install a handler in the current DB."
         self.lastDbProgress = 0
         self.inDB = False
-        try:
-            db.set_progress_handler(self._dbProgress, 10000)
-        except:
-            print """\
-Your pysqlite2 is too old. Anki will appear frozen during long operations."""
+        db.set_progress_handler(self._dbProgress, 10000)
 
     def _dbProgress(self):
         "Called from SQLite."
@@ -51,87 +49,116 @@ Your pysqlite2 is too old. Anki will appear frozen during long operations."""
         self.inDB = True
         # handle GUI events
         if not self.blockUpdates:
-          self._maybeShow()
-          self.app.processEvents(QEventLoop.ExcludeUserInputEvents)
+            self._maybeShow()
+            self.app.processEvents(QEventLoop.ExcludeUserInputEvents)
         self.inDB = False
 
-    # DB-safe timers
+    # Safer timers
     ##########################################################################
     # QTimer may fire in processEvents(). We provide a custom timer which
-    # automatically defers until the DB is not busy.
+    # automatically defers until the DB is not busy, and avoids running
+    # while a progress window is visible.
 
-    def timer(self, ms, func, repeat):
+    def timer(self, ms, func, repeat, requiresCollection=True):
         def handler():
-            if self.inDB:
+            if self.inDB or self._levels:
                 # retry in 100ms
-                self.timer(100, func, repeat)
+                self.timer(100, func, False, requiresCollection)
+            elif not self.mw.col and requiresCollection:
+                # ignore timer events that fire after collection has been
+                # unloaded
+                print("Ignored progress func as collection unloaded: %s" % repr(func))
             else:
                 func()
         t = QTimer(self.mw)
         if not repeat:
             t.setSingleShot(True)
-        t.connect(t, SIGNAL("timeout()"), handler)
+        t.timeout.connect(handler)
         t.start(ms)
         return t
 
     # Creating progress dialogs
     ##########################################################################
 
-    class ProgressNoCancel(QProgressDialog):
+    class ProgressDialog(QDialog):
+        def __init__(self, parent):
+            QDialog.__init__(self, parent)
+            self.form = aqt.forms.progress.Ui_Dialog()
+            self.form.setupUi(self)
+            self._closingDown = False
+            self.wantCancel = False
+
+        def cancel(self):
+            self._closingDown = True
+            self.hide()
+
         def closeEvent(self, evt):
-            evt.ignore()
+            if self._closingDown:
+                evt.accept()
+            else:
+                self.wantCancel  = True
+                evt.ignore()
+
         def keyPressEvent(self, evt):
             if evt.key() == Qt.Key_Escape:
                 evt.ignore()
+                self.wantCancel  = True
 
     def start(self, max=0, min=0, label=None, parent=None, immediate=False):
         self._levels += 1
         if self._levels > 1:
             return
         # setup window
-        parent = parent or self.app.activeWindow() or self.mw
+        parent = parent or self.app.activeWindow()
+        if not parent and self.mw.isVisible():
+            parent = self.mw
+
         label = label or _("Processing...")
-        self._win = self.ProgressNoCancel(label, "", min, max, parent)
+        self._win = self.ProgressDialog(parent)
+        self._win.form.progressBar.setMinimum(min)
+        self._win.form.progressBar.setMaximum(max)
+        self._win.form.progressBar.setTextVisible(False)
+        self._win.form.label.setText(label)
         self._win.setWindowTitle("Anki")
-        self._win.setCancelButton(None)
-        self._win.setAutoClose(False)
-        self._win.setAutoReset(False)
         self._win.setWindowModality(Qt.ApplicationModal)
-        # we need to manually manage minimum time to show, as qt gets confused
-        # by the db handler
-        self._win.setMinimumDuration(100000)
+        self._win.setMinimumWidth(300)
         if immediate:
-            self._shown = True
-            self._win.show()
-            self.app.processEvents()
+            self._showWin()
         else:
             self._shown = False
         self._counter = min
         self._min = min
         self._max = max
         self._firstTime = time.time()
-        self._lastTime = time.time()
-        self._disabled = False
+        self._lastUpdate = time.time()
+        self._updating = False
+        return self._win
 
     def update(self, label=None, value=None, process=True, maybeShow=True):
         #print self._min, self._counter, self._max, label, time.time() - self._lastTime
+        if self._updating:
+            return
         if maybeShow:
             self._maybeShow()
-        self._lastTime = time.time()
+        if not self._shown:
+            return
+        elapsed = time.time() - self._lastUpdate
         if label:
-            self._win.setLabelText(label)
-        if self._max and self._shown:
+            self._win.form.label.setText(label)
+        if self._max:
             self._counter = value or (self._counter+1)
-            self._win.setValue(self._counter)
-        if process:
+            self._win.form.progressBar.setValue(self._counter)
+        if process and elapsed >= 0.2:
+            self._updating = True
             self.app.processEvents(QEventLoop.ExcludeUserInputEvents)
+            self._updating = False
+            self._lastUpdate = time.time()
 
     def finish(self):
         self._levels -= 1
         self._levels = max(0, self._levels)
         if self._levels == 0 and self._win:
-            self._win.cancel()
-            self._unsetBusy()
+            self._closeWin()
 
     def clear(self):
         "Restore the interface after an error."
@@ -147,16 +174,33 @@ Your pysqlite2 is too old. Anki will appear frozen during long operations."""
             return
         delta = time.time() - self._firstTime
         if delta > 0.5:
-            self._shown = True
-            self._win.show()
-            self._setBusy()
+            self._showWin()
+
+    def _showWin(self):
+        self._shown = time.time()
+        self._win.show()
+        self._setBusy()
+
+    def _closeWin(self):
+        if self._shown:
+            while True:
+                # give the window system a second to present
+                # window before we close it again - fixes
+                # progress window getting stuck, especially
+                # on ubuntu 16.10+
+                elap = time.time() - self._shown
+                if elap >= 0.5:
+                    break
+                self.app.processEvents(QEventLoop.ExcludeUserInputEvents)
+        self._win.cancel()
+        self._win = None
+        self._shown = False
+        self._unsetBusy()
 
     def _setBusy(self):
-        self._disabled = True
         self.mw.app.setOverrideCursor(QCursor(Qt.WaitCursor))
 
     def _unsetBusy(self):
-        self._disabled = False
         self.app.restoreOverrideCursor()
 
     def busy(self):
